#include <crypto/hash.h>
#include <linux/workqueue.h>
#include <linux/radix-tree.h>
#include <linux/rcupdate.h>
#include "sdfs_common.h"
#include "ext4.h"
#include "sdfs_store.h"
#include "sdfs_filter.h"
#include "sha1_ssse3_glue_em.c"

void ldfs_prev_init(struct super_block * sb) {
	EXT4_SB(sb)->s_blk_root = RB_ROOT;
}

int ldfs_prev_exist(struct super_block * sb, __u8 fp[]) {
	struct rb_root * root = &EXT4_SB(sb)->s_blk_root;
	struct rb_node * node = root->rb_node;

	while (node) {
		struct ldfs_blk_info * info = container_of(node, struct ldfs_blk_info, node);
		int result = memcmp(fp, info->new_fp, EXT4_SB(sb)->s_fp_info->len);
		if (result < 0)
			node = node->rb_left;
		else if (result > 0)
			node = node->rb_right;
		else
			return 1;
	}
	return 0;
}

int ldfs_prev_insert(struct super_block * sb, struct ldfs_blk_info * i) {
	struct rb_root * root = &EXT4_SB(sb)->s_blk_root;
	struct rb_node **new = &(root->rb_node), *parent = NULL;
	struct ldfs_blk_info * info;
	int result;

	/* Figure out where to put new node */
	while (*new) {
		info = container_of(*new, struct ldfs_blk_info, node);
		result = memcmp(i->new_fp, info->new_fp, EXT4_SB(sb)->s_fp_info->len);

		parent = *new;
		if (result < 0)
			new = &((*new)->rb_left);
		else if (result > 0)
			new = &((*new)->rb_right);
		else
			return 0;
	}

	/* Add new node and rebalance tree. */
	rb_link_node(&i->node, parent, new);
	rb_insert_color(&i->node, root);
	return 1;
}

int crypto_shash_sha1_ssse3_update(struct shash_desc *shash, const u8 *data, unsigned int len) {
	int ret;
	if ((ret = sha1_ssse3_update(shash, data, len)) == -1) {
		ret = crypto_shash_update(shash, data, len);
	}
	return ret;
}

struct ldfs_fp_info md5_info = {
	.name = "md5",
	.len = 16,
	.zero_result = "\x62\x0f\x0b\x67\xa9\x1f\x7f\x74\x15\x1b\xc5\xbe\x74\x5b\x71\x10",
	.update = crypto_shash_update,
};

struct ldfs_fp_info sha1_info = {
	.name = "sha1",
	.len = 20,
	.zero_result = "\x1c\xea\xf7\x3d\xf4\x0e\x53\x1d\xf3\xbf\xb2\x6b\x4f\xb7\xcd\x95\xfb\x7b\xff\x1d",
	.update = crypto_shash_update,
};

struct ldfs_fp_info sha1_ssse3_info = {
	.name = "sha1",
	.len = 20,
	.zero_result = "\x1c\xea\xf7\x3d\xf4\x0e\x53\x1d\xf3\xbf\xb2\x6b\x4f\xb7\xcd\x95\xfb\x7b\xff\x1d",
	.update = crypto_shash_sha1_ssse3_update,
};

struct ldfs_fp_info tgr_info = {
	.name = "tgr192",
	.len = 24,
	.zero_result = "\xc5\x0b\x60\xfa\x71\x0a\xa9\x4f\x3e\x56\x15\xb4\xd4\xa1\x94\xe3\xcd\xfa\xcf\x95\xe0\x25\x3b\x6a",
	.update = crypto_shash_update,
};

/*
 *	default implementation is sha1
 */
int ldfs_fp_info_assign(struct super_block * sb) {
	EXT4_SB(sb)->s_fp_info = &tgr_info;
	return 0;
}

static int __ldfs_fp_compute(struct super_block * sb, __u8 dst[], char * src, int len, struct sdesc *sdesc) {
	crypto_shash_init(&sdesc->shash);
	EXT4_SB(sb)->s_fp_info->update(&sdesc->shash, src, len);
	crypto_shash_final(&sdesc->shash, dst);
	return 0;
}

int ldfs_fp_compute(struct super_block * sb, __u8 dst[], char * src, int len) {
	struct crypto_shash * tfm;
	struct sdesc * sdesc;
	unsigned int size;
	tfm = crypto_alloc_shash(EXT4_SB(sb)->s_fp_info->name, 0, 0);
	size = sizeof(struct shash_desc) + crypto_shash_descsize(tfm);
	sdesc = kmalloc(size, GFP_KERNEL);
	sdesc->shash.tfm = tfm;
	sdesc->shash.flags = 0;
	crypto_shash_init(&sdesc->shash);
	crypto_shash_update(&sdesc->shash, src, len);
	crypto_shash_final(&sdesc->shash, dst);
	crypto_free_shash(sdesc->shash.tfm);
	kfree(sdesc);
	return 0;
}

int ldfs_fp_equal(struct super_block * sb, __u8 * fp1, __u8 * fp2) {
	return (!memcmp(fp1, fp2, EXT4_SB(sb)->s_fp_info->len));
}

int ldfs_fp_zero(struct super_block * sb, __u8 * fp) {
	return ldfs_fp_equal(sb, fp, EXT4_SB(sb)->s_fp_info->zero_result);
}

void ldfs_fp_print(struct super_block * sb, __u8 * fp) {
	int i;
	for(i = 0; i <  EXT4_SB(sb)->s_fp_info->len; i++) {
		printk("%02x", fp[i]);
	}
	printk("\n");
}

int ldfs_workqueue_init(struct super_block * sb,int thread_no) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
    int i = 0;
    sbi->s_fingerprint_wq = alloc_workqueue("LDFS_FP_WQ", WQ_NON_REENTRANT | WQ_UNBOUND, 0);
    sbi->s_ph.ph_sb = sb;
    sbi->s_ph.ph_jobs= vmalloc(sizeof(struct ldfs_job_info) * thread_no);
    sbi->s_ph.ph_thread_no = thread_no;
    mutex_init(&sbi->s_ph.ph_lock);
    for(i = 0; i < thread_no; i++){
        struct crypto_shash * tfm;
        struct sdesc * sdesc;
        unsigned int size;
        tfm = crypto_alloc_shash(EXT4_SB(sb)->s_fp_info->name, 0, 0);
        size = sizeof(struct shash_desc) + crypto_shash_descsize(tfm);
        sdesc = kmalloc(size, GFP_KERNEL);
        sdesc->shash.tfm = tfm;
        sdesc->shash.flags = 0;
        sbi->s_ph.ph_jobs[i].sdesc = sdesc;
        mutex_init(&sbi->s_ph.ph_jobs[i].work_mutex);
        sbi->s_ph.ph_jobs[i].blk_data = sbi->s_blk_info;
    }

	return 0;
}

void ldfs_workqueue_exit(struct super_block * sb) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
    int i ;
    int thread_no = sbi->s_ph.ph_thread_no;
    for(i = 0; i < thread_no; i++){
        struct sdesc * sdesc = sbi->s_ph.ph_jobs[i].sdesc;
        crypto_free_shash(sdesc->shash.tfm);
        kfree(sdesc);
    }
    vfree(sbi->s_ph.ph_jobs);
	destroy_workqueue(sbi->s_fingerprint_wq);
}

static void do_fingerprint_work_first_part(struct ldfs_blk_info * b_info, struct sdesc * sdesc) {
	struct inode *inode = b_info->page->mapping->host;
	struct super_block * sb = inode->i_sb;

	// TODO : SHOULD NOT BE PAGE_SIZE!
	__ldfs_fp_compute(sb, b_info->new_fp, b_info->bh->b_data, PAGE_SIZE, sdesc);
	b_info->state = ldfs_fingerprint_search(sb, b_info->new_fp, &b_info->new_blocknr);
}

static void do_fingerprint_work(struct work_struct * work) {
	struct ldfs_job_info * info = container_of(work, struct ldfs_job_info, page_work);
	int i = 0;
	for (i = 0; i < info->max_len; i++) {
		if (atomic_dec_and_test(&info->blk_data[i].fp_finished)) {
			do_fingerprint_work_first_part(&info->blk_data[i], info->sdesc);
			mutex_unlock(&info->blk_data[i].fp_finished_mutex);
		}
	}
	mutex_unlock(&info->work_mutex);
}

int ldfs_fingerprint_enqueue(struct super_block * sb, struct ldfs_job_info * info, int cpu) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
	INIT_WORK(&info->page_work, do_fingerprint_work);
	mutex_lock(&info->work_mutex);
	queue_work_on(cpu, sbi->s_fingerprint_wq, &info->page_work);
	return 0;
}

void ldfs_fingerprint_compute_batch(struct super_block * sb, int len, struct ldfs_blk_info * b_info) {
	int cur_cpu = get_cpu();
	struct ext4_sb_info * sbi = EXT4_SB(sb);
	int thread_no = sbi->s_ph.ph_thread_no;
	int cpu = 0;
	int i = 0, j = 0;
	int max_cpu = num_online_cpus();

	for (cpu = 0; cpu < max_cpu; cpu++) {
		if (cpu != cur_cpu || max_cpu == 1) {
			sbi->s_ph.ph_jobs[i].max_len = len;
			ldfs_fingerprint_enqueue(sb, &sbi->s_ph.ph_jobs[i++], cpu);
			if (--thread_no == 0) {
				break;
			}
		}
	}
	for(j = 0; j < i; j++){
		mutex_lock(&sbi->s_ph.ph_jobs[j].work_mutex);
		mutex_unlock(&sbi->s_ph.ph_jobs[j].work_mutex);
	}

}
