#include <linux/log2.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/crypto.h>
#include "sdfs_common.h"
#include "sdfs_filter.h"
#include "sdfs_store.h"
#include "sdfs_lcache.h"

__u32   __ldfs_index_key(struct super_block * sb, __u8 hash[])   {
struct ext4_sb_info * sbi = EXT4_SB(sb);
__u32 key = *(__u32 *)(hash + 4);
return key >> (32 - sbi->s_index_bit);
}

static inline __u32 ldfs_part_fp(__u8 fp[]) {
	return fp[0] << 16 | fp[1] << 8 | fp[2];
}

static inline __u32 ldfs_part_fp_get(struct ldfs_part_fp part_fp) {
	return part_fp.data[0] << 16 | part_fp.data[1] << 8 | part_fp.data[2];
}

static inline void ldfs_part_fp_set(struct ldfs_part_fp * part_fp, __u8 full_fp[]) {
	memcpy(part_fp, full_fp, 3);
}

struct ldfs_bucket * ldfs_bucket_alloc(struct super_block * sb, struct slist_head * head) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
	struct ldfs_bucket * bucket = kmem_cache_alloc(sbi->s_bucket_cache, GFP_KERNEL);
	memset(bucket, 0, LDFS_BUCKET_SIZE);
	slist_add(&bucket->b_head, head);
	sbi->s_bucket_count++;
	return bucket;
}

int ldfs_bucket_free(struct super_block * sb, struct slist_head * head, struct ldfs_bucket * bucket) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
	slist_del(&bucket->b_head, head);
	kmem_cache_free(sbi->s_bucket_cache, bucket);
	sbi->s_bucket_count--;
	return 0;
}

int ldfs_filter_bucket_search(struct super_block * sb, struct ldfs_bucket * bucket, __u8 fp[], ext4_fsblk_t * blocknr) {
	int left = 0, right = bucket->b_count, middle = 0;
	__u32 this_p_fp = ldfs_part_fp(fp);
	__u32 bucket_p_fp;
	int i;

	// Binary Search
	while (left <= right) {
		middle = (left + right) / 2;
		bucket_p_fp = ldfs_part_fp_get(bucket->b_part_fp[middle]);
		if (this_p_fp > bucket_p_fp) {
			left = middle + 1;
		} else if (this_p_fp < bucket_p_fp) {
			right = middle - 1;
		} else {
			while (middle > 0 && this_p_fp == ldfs_part_fp_get(bucket->b_part_fp[middle - 1])) {
				middle--;
			}
			for (i = middle; i < bucket->b_count && this_p_fp == ldfs_part_fp_get(bucket->b_part_fp[i]); i++) {
#ifndef LDFS_FILTER_TEST
				if (!ldfs_lcache_search_group(sb, fp, bucket->b_part_blocknr[i], blocknr)) {
					return LDFS_DUPL;
				}
#else
				__u8 comp_fp[LDFS_FP_LEN_MAX];
				ldfs_store_read(sb, bucket->b_part_blocknr[i], comp_fp, NULL);
				if(ldfs_fp_equal(sb, fp, comp_fp)) {
					*blocknr = bucket->b_part_blocknr[i];
					return LDFS_DUPL;
				}
#endif
			}
			break;
		}
	}
	return LDFS_UNIQUE;
}

int ldfs_filter_bucket_insert(struct super_block * sb, struct ldfs_bucket * bucket, __u8 fp[], ext4_fsblk_t blocknr) {
	int left = 0, right = bucket->b_count - 1, middle = 0;
	__u32 this_p_fp = ldfs_part_fp(fp);
	__u32 bucket_p_fp;

	// Binary Search
	while (left <= right) {
		middle = (left + right) / 2;
		bucket_p_fp = ldfs_part_fp_get(bucket->b_part_fp[middle]);
		if (this_p_fp > bucket_p_fp) {
			left = middle + 1;
		} else if (this_p_fp < bucket_p_fp) {
			right = middle - 1;
		} else {
			break;
		}
	}
	if (left == middle + 1) {
		middle++;
	}
#ifndef LDFS_FILTER_TEST
	memmove(&bucket->b_part_blocknr[middle + 1], &bucket->b_part_blocknr[middle],
			(bucket->b_count - middle) * sizeof(__u16 ));
	bucket->b_part_blocknr[middle] = (blocknr / LDFS_STORE_ENTRIES(sb));
#else
	memmove(&bucket->b_part_blocknr[middle + 1], &bucket->b_part_blocknr[middle],
			(bucket->b_count - middle) * sizeof(__u32 ));
	bucket->b_part_blocknr[middle] = blocknr;
#endif
	memmove(&bucket->b_part_fp[middle + 1], &bucket->b_part_fp[middle],
			(bucket->b_count - middle) * sizeof(struct ldfs_part_fp));
	ldfs_part_fp_set(&bucket->b_part_fp[middle], fp);

	bucket->b_count++;
	return 0;
}

int ldfs_filter_bucket_delete(struct super_block * sb, struct ldfs_bucket * bucket, __u8 fp[], ext4_fsblk_t blocknr) {
	int left = 0, right = bucket->b_count - 1, middle = 0;
	int ret = -1;
	int i;
	int hit = 0;
	__u32 this_p_fp = ldfs_part_fp(fp);
	__u32 bucket_p_fp;

	if (bucket->b_count == 0) {
		return -1;
	}
	// Binary Search
	while (left <= right) {
		middle = (left + right) / 2;
		bucket_p_fp = ldfs_part_fp_get(bucket->b_part_fp[middle]);
		if (this_p_fp > bucket_p_fp) {
			left = middle + 1;
		} else if (this_p_fp < bucket_p_fp) {
			right = middle - 1;
		} else {
			while (middle > 0 && this_p_fp == ldfs_part_fp_get(bucket->b_part_fp[middle - 1])) {
				middle--;
			}
			hit = 1;
			break;
		}
	}
	if (!hit) {
		return ret;
	}
	for (i = middle; i < bucket->b_count && this_p_fp == ldfs_part_fp_get(bucket->b_part_fp[i]); i++) {
#ifndef LDFS_FILTER_TEST
		__u16 storenr = blocknr / LDFS_STORE_ENTRIES(sb);
		if (storenr == bucket->b_part_blocknr[i]) {
			memmove(&bucket->b_part_blocknr[i], &bucket->b_part_blocknr[i + 1],
					(bucket->b_count - (i + 1)) * sizeof(__u16 ));
			memmove(&bucket->b_part_fp[i], &bucket->b_part_fp[i + 1],
					(bucket->b_count - (i + 1)) * sizeof(struct ldfs_part_fp));
			bucket->b_count--;
			ret = 0;
			break;
		}
#else
		if (blocknr == bucket->b_part_blocknr[i]) {
			memmove(&bucket->b_part_blocknr[i], &bucket->b_part_blocknr[i + 1],
					(bucket->b_count - (i + 1)) * sizeof(__u32 ));
			memmove(&bucket->b_part_fp[i], &bucket->b_part_fp[i + 1],
					(bucket->b_count - (i + 1)) * sizeof(struct ldfs_part_fp));
			bucket->b_count--;
			ret = 0;
			break;
		}
#endif
	}
	return ret;
}

int ldfs_filter_init(struct super_block * sb) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
	unsigned index_head_cnt, i;

	mutex_init(&sbi->s_prefetch_mutex);

	// Init the filter cache
	sbi->s_bucket_count = 0;
	if (!sbi->s_index_bit) {
		sbi->s_index_bit = ilog2(ext4_blocks_count(sbi->s_es)) - 8;
	}
	index_head_cnt = (1 << sbi->s_index_bit);
	sbi->s_bucket_head = vmalloc(index_head_cnt * sizeof(struct slist_head));
	sbi->s_bucket_cache = kmem_cache_create("ldfs_bucket_cache", LDFS_BUCKET_SIZE, 0, 0, NULL);

	for (i = 0; i < index_head_cnt; i++) {
		slist_head_init(&sbi->s_bucket_head[i]);
	}

	printk(KERN_INFO "LiveDFS Index_Key Bits = %u", sbi->s_index_bit);

	ldfs_store_mount(sb);
	ldfs_lcache_init(sb);
	return 0;
}

int ldfs_filter_exit(struct super_block * sb) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
	struct slist_head * head;
	struct ldfs_bucket * bucket;
	int i;

	ldfs_lcache_exit(sb);

	for (i = 0; i < (1 << sbi->s_index_bit); i++) {
		head = &sbi->s_bucket_head[i];
		while (!slist_empty(head)) {
			bucket = slist_first_entry(head, struct ldfs_bucket, b_head);
			ldfs_bucket_free(sb, head, bucket);
		}
	}
	kmem_cache_destroy(sbi->s_bucket_cache);
	vfree(sbi->s_bucket_head);
	return 0;
}

int ldfs_filter_search(struct super_block * sb, __u8 fp[], ext4_fsblk_t * blocknr) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
	__u32 index = __ldfs_index_key(sb, fp);
	struct slist_head * head;
	struct ldfs_bucket * bucket;
	int ret = LDFS_UNIQUE;

	slist_for_each(head, &sbi->s_bucket_head[index]) {
		bucket = slist_entry(head, struct ldfs_bucket, b_head);
		ret = ldfs_filter_bucket_search(sb, bucket, fp, blocknr);
		if (ret == LDFS_DUPL) {
			break; // Entry Found
		}
	}
	return ret;
}

int ldfs_filter_insert_entry(struct super_block * sb, __u8 fp[], ext4_fsblk_t blocknr) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
	__u32 index = __ldfs_index_key(sb, fp);
	struct slist_head * head;
	struct ldfs_bucket * bucket;

	slist_for_each(head, &sbi->s_bucket_head[index]) {
		bucket = slist_entry(head, struct ldfs_bucket, b_head);
		if (bucket->b_count < LDFS_BUCKET_ENTRY_MAX) {
			return ldfs_filter_bucket_insert(sb, bucket, fp, blocknr);
		}
	}
	bucket = ldfs_bucket_alloc(sb, &sbi->s_bucket_head[index]);
	return ldfs_filter_bucket_insert(sb, bucket, fp, blocknr);
}

int ldfs_filter_remove_entry(struct super_block * sb, __u8 fp[], ext4_fsblk_t blocknr) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
	__u32 index = __ldfs_index_key(sb, fp);
	struct slist_head * head;
	struct ldfs_bucket * bucket;
	int ret = -1;

	slist_for_each(head, &sbi->s_bucket_head[index]) {
		bucket = slist_entry(head, struct ldfs_bucket, b_head);
		ret = ldfs_filter_bucket_delete(sb, bucket, fp, blocknr);
		if (!ret) {
			if (!bucket->b_count) {
				ldfs_bucket_free(sb, &sbi->s_bucket_head[index], bucket);
			}
			return ret;
		}
	}
	return ret;
}

int ldfs_fingerprint_search(struct super_block * sb, __u8 fp[], ext4_fsblk_t * blocknr) {
	if (ldfs_fp_zero(sb, fp)) {
		*blocknr = 0;
		return LDFS_ZERO;
	}

	return ldfs_filter_search(sb, fp, blocknr);
}

int ldfs_fingerprint_insert(struct super_block * sb, __u8 fp[], ext4_fsblk_t blocknr) {
	if (ldfs_fp_zero(sb, fp)) {
		return 0;
	}
	ldfs_filter_insert_entry(sb, fp, blocknr);
	ldfs_lcache_invalidate(sb, blocknr);
	return 0;
}

int ldfs_fingerprint_remove(struct super_block * sb, __u8 fp[], ext4_fsblk_t blocknr) {
	if (ldfs_fp_zero(sb, fp)) {
		return 0;
	}
	ldfs_filter_remove_entry(sb, fp, blocknr);
	ldfs_lcache_invalidate(sb, blocknr);
	return 0;
}

