/*
 * sdfs_read.c
 *
 *  Created on: Nov 19, 2012
 *      Author: ngchunho
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/list.h>
#include <linux/string.h>
#include <linux/mpage.h>
#include <linux/sort.h>

#include "ext4.h"

#include "sdfs_store.h"

#define SORT_SIZE 2048

struct block_page {
	ext4_fsblk_t blocknr;
	struct page * page;
};

static int cmp_func(const void * a, const void * b) {
	struct block_page * aa, *bb;
	aa = (struct block_page *) a;
	bb = (struct block_page *) b;
	if (aa->blocknr > bb->blocknr) {
		return 1;
	} else if (aa->blocknr < bb->blocknr) {
		return -1;
	}
	return 0;
}

static void block_page_swap(void * aa, void * bb,
		int size) {
	struct block_page * a = aa;
	struct block_page * b = bb;
	struct block_page tmp_page;

	tmp_page = *a;
	*a = *b;
	*b = tmp_page;
}

static struct ldfs_read_cache_page *
ldfs_read_cache_search(struct ext4_sb_info * sbi, unsigned b_blocknr) {
	struct ldfs_read_cache_page * read_page;
	HASH_FIND_INT(sbi->s_read_cache_hash, &b_blocknr, read_page);
	return read_page;
}
static int
ldfs_read_cache_insert(struct ext4_sb_info *sbi,
		struct ldfs_read_cache_page *data) {
	HASH_ADD_INT(sbi->s_read_cache_hash, page->index, data);
	return 1;
}

static void ldfs_process_read_cache(struct page * page, ext4_fsblk_t thisB) {
	lock_page(page);
	do {
		ext4_fsblk_t b_blocknr = thisB;
		struct inode *inode = page->mapping->host;
		struct super_block * sb = inode->i_sb;
		struct ext4_sb_info * sbi = EXT4_SB(sb);
		struct ldfs_read_cache_page * cache_page;
		unsigned need_cache = 1;

		mutex_lock(&sbi->s_read_cache_mutex);

		cache_page = ldfs_read_cache_search(sbi, b_blocknr);

		if (cache_page) {
		} else {
			unsigned new_blk_count;
			unsigned last_blk_no;
			ldfs_store_read(sb, b_blocknr, NULL, &new_blk_count);
			if (new_blk_count > 1) {
				cache_page = list_first_entry(&sbi->s_read_cache_list,
						struct ldfs_read_cache_page,lru);
				lock_page(cache_page->page);

				last_blk_no = cache_page->page->index;

				if (last_blk_no) {
					HASH_DEL(sbi->s_read_cache_hash, cache_page);
				}

				cache_page->page->index = b_blocknr;

				if (ldfs_read_cache_insert(sbi, cache_page) == 0) {
					printk(KERN_EMERG "last_blk_no = %u", last_blk_no);

					BUG_ON(1);
				}

				//printk(KERN_INFO "Insert %u",b_blocknr);

				memcpy(kmap(cache_page->page), kmap(page), PAGE_SIZE);
				kunmap(cache_page->page);
				kunmap(page);
			} else
				need_cache = 0;
		}

		if (need_cache
				&& !list_is_last(&cache_page->lru, &sbi->s_read_cache_list)) {
			list_move_tail(&cache_page->lru, &sbi->s_read_cache_list);
			//Anyway we need to move tmp to tail
		}
		if (cache_page) {
			unlock_page(cache_page->page);
		}
		mutex_unlock(&sbi->s_read_cache_mutex);

	} while (0);

	unlock_page(page);
}
int sdfs_readpages(struct file *file, struct address_space *mapping,
		struct list_head *pages, unsigned nr_pages) {
	unsigned page_idx;
	//   struct list_head pages_batch;
	struct inode * inode = mapping->host;
	struct buffer_head bh_result;
	ext4_fsblk_t seek1 = 0, lastbk = 0, zeros = 0;
	int blocks_to_boundary = 10000;
	//struct ext4_sb_info * sbi = EXT4_SB(inode->i_sb);
	struct block_page * block_n_pages;
	block_n_pages = vmalloc(SORT_SIZE * sizeof(struct block_page));

	if (nr_pages > SORT_SIZE) {
		printk(KERN_INFO "nr_pages = %u",nr_pages);
		BUG_ON(nr_pages > SORT_SIZE);
	}

	for (page_idx = 0; page_idx < nr_pages; page_idx++) {
		struct page
		*page = list_entry(pages->prev, struct page, lru);
		ext4_fsblk_t thisB;

		list_del(&page->lru);
		if (page_idx < blocks_to_boundary) {
			bh_result.b_state = 0;
			bh_result.b_page = page;
			bh_result.b_size = (1) << inode->i_blkbits;
			ext4_get_block(inode, page->index, &bh_result, 0);
			if (buffer_mapped(&bh_result)) {
				block_n_pages[page_idx].page = page;
				block_n_pages[page_idx].blocknr = bh_result.b_blocknr;
				thisB = block_n_pages[page_idx].blocknr;
				//printk(KERN_INFO "ino = %u,%lu is mapped,blockno = %u",inode->i_ino,page->index,bh_result.b_blocknr);
			} else {
				//printk(KERN_INFO "%lu is not mapped",page->index);
				block_n_pages[page_idx].page = page;
				block_n_pages[page_idx].blocknr = 0;
				thisB = 0;
				zeros++;
			}
			if (thisB != 0 && thisB != lastbk + 1 && thisB != lastbk) {
				seek1++;
			}
			lastbk = thisB;
		} else
			page_cache_release(page);
	}

	if (nr_pages > blocks_to_boundary)
		nr_pages = blocks_to_boundary;

//    printk(KERN_EMERG "SORTING");
#ifdef LDFS_READ_REORDER_ENABLE
	if (seek1 > 1)
		sort(block_n_pages, nr_pages, sizeof(struct block_page), cmp_func,
				block_page_swap);
#endif

	//   printk(KERN_EMERG "SORTED");
	lastbk = 0;
	for (page_idx = 0; page_idx < nr_pages; page_idx++) {
		struct page *page = block_n_pages[page_idx].page;
		ext4_fsblk_t thisB = block_n_pages[page_idx].blocknr;

		add_to_page_cache_lru(page, mapping, page->index, GFP_KERNEL);
		//      printk(KERN_EMERG "page index before readpage = %u",page->index);

		if (thisB == 0) {
			memset(kmap(page), 0, PAGE_SIZE);
			kunmap(page);
			SetPageUptodate(page);
		} else if (thisB == lastbk && page_idx != 0) {
		} else {
			struct ext4_sb_info *sbi = EXT4_SB(inode->i_sb);
#ifdef LDFS_READ_CACHE_ENABLE
			struct ldfs_read_cache_page * cache;
			mutex_lock(&sbi->s_read_cache_mutex);
			cache = ldfs_read_cache_search(
					sbi,
					thisB);

			if(cache) {
				//4K block size
				memcpy(kmap(page),kmap(cache->page),PAGE_SIZE);
				kunmap(page);
				kunmap(cache->page);
				SetPageUptodate(page);
				mutex_unlock(&sbi->s_read_cache_mutex);
			}
			else {
				mutex_unlock(&sbi->s_read_cache_mutex);
				// printk(KERN_EMERG "page index readpage = %u",page->index);
				mpage_readpage(page, ext4_get_block);
			}
#else
			mpage_readpage(page, ext4_get_block);
#endif
		}
		lastbk = thisB;
	}

	//printk(KERN_EMERG "SUBMITTED");
	lastbk = 0;
	for (page_idx = 0; page_idx < nr_pages; page_idx++) {
		struct page *page = block_n_pages[page_idx].page;
		ext4_fsblk_t thisB = block_n_pages[page_idx].blocknr;

		if (thisB == lastbk && page_idx != 0 && thisB != 0) {
			if (!PageUptodate(block_n_pages[page_idx].page)) {
				/*
				 * This is bad way to avoid read same block many times.
				 * Because we are busy waiting the I/O, which is break
				 * the protocal that readpages should be async.
				 *
				 * I need to write my own readpage function instead of
				 * using mpage_readpage. Which could handle this when
				 * I/O complete is called.
				 */
				//           printk(KERN_EMERG "wating page = %u",block_n_pages[page_idx - 1].page->index);
				while (!PageUptodate(block_n_pages[page_idx - 1].page)) {
					io_schedule();
				}
				//          printk(KERN_EMERG "end wating page = %u",block_n_pages[page_idx - 1].page->index);

				memcpy(kmap(block_n_pages[page_idx - 1].page), kmap(page),
						PAGE_SIZE);

				kunmap(block_n_pages[page_idx - 1].page);
				kunmap(page);
				SetPageUptodate(page);
			}
		}
		unlock_page(page);
		lastbk = thisB;
	}
	//printk(KERN_EMERG "Before process");

	lastbk = 0;
	for (page_idx = 0; page_idx < nr_pages; page_idx++) {
		struct page *page = block_n_pages[page_idx].page;
		ext4_fsblk_t thisB = block_n_pages[page_idx].blocknr;

#ifdef LDFS_READ_CACHE_ENABLE
		if(thisB != 0 && lastbk != thisB) {
			//printk(KERN_EMERG "SORTED page->index = %u",page->index);
			while(!PageUptodate(page)) {
				io_schedule();
			}
			//printk(KERN_EMERG "Updated page->index = %u",page->index);
			ldfs_process_read_cache(page,thisB);
			//printk(KERN_EMERG "Processed page->index = %u",page->index);
		}
#endif
		page_cache_release(page);

		lastbk = thisB;
	}

	/*
	 printk(KERN_INFO "seek1 = %lu,seek2 = %lu,zeros = %lu,nr_pages=%u,hit=%u",
	 seek1,seek2,zeros,nr_pages,sbi->s_read_cache_hit_count);
	 */

	vfree(block_n_pages);
	return 0;
	//return mpage_readpages(mapping, pages, nr_pages, ext3_get_block);
}
