#include <linux/slab.h>
#include <linux/vmalloc.h>
#include "ext4_jbd2.h"
#include "sdfs_store.h"
#include "sdfs_filter.h"
#include "sdfs_common.h"
#include "sdfs_lcache.h"


static inline __u32 ldfs_store_block_offset(ext4_grpblk_t offsetp) {
	return offsetp % LDFS_S_CNT;
}

static inline ext4_fsblk_t ldfs_store_blocknr(struct super_block * sb, ext4_group_t blockgrpp, ext4_grpblk_t offsetp) {
	ext4_group_t group = blockgrpp & (~(LDFS_STORE_COMPACT - 1));
	struct ext4_group_desc * desc = ext4_get_group_desc(sb, group, NULL);
	__u32 offset = (blockgrpp & (LDFS_STORE_COMPACT - 1)) * EXT4_BLOCKS_PER_GROUP(sb) + offsetp;
	return desc->bg_block_store + offset / LDFS_S_CNT;
}

int ldfs_store_prefetched(struct super_block * sb, ext4_fsblk_t blocknr) {
	struct buffer_head * bh;
	bh = __find_get_block(sb->s_bdev, blocknr, PAGE_SIZE);
	if (!bh) {
		return 0;
	}
	brelse(bh);
	return 1;
}

int ldfs_store_prefetch(struct super_block * sb, ext4_group_t group) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
	ext4_fsblk_t store_blocknr = ldfs_store_blocknr(sb, group, 0);
	__u32 store_size = EXT4_BLOCKS_PER_GROUP(sb) / LDFS_S_CNT * LDFS_STORE_COMPACT;
	char * load = kzalloc(store_size, GFP_KERNEL);
	struct buffer_head ** bh = kmalloc(store_size * sizeof(struct buffer_head *), GFP_KERNEL);
	int i;

	mutex_lock(&sbi->s_prefetch_mutex);
	for (i = 0; i < store_size; i++) {
		bh[i] = sb_getblk(sb, store_blocknr + i);
		if (buffer_uptodate(bh[i])) {
			brelse(bh[i]);
			continue;
		}
		load[i] = 1;
		lock_buffer(bh[i]);
		get_bh(bh[i]);
		bh[i]->b_end_io = end_buffer_read_sync;
		submit_bh(READ, bh[i]);
	}
	for (i = 0; i < store_size; i++) {
		if (load[i]) {
			wait_on_buffer(bh[i]);
			BUG_ON(!buffer_uptodate(bh[i]));
			brelse(bh[i]);
		}
	}
	mutex_unlock(&sbi->s_prefetch_mutex);
	kfree(bh);
	kfree(load);
	return 0;
}

ext4_fsblk_t ldfs_get_store_position(struct super_block * sb, ext4_fsblk_t blocknr, __u32 * offset) {
	ext4_group_t blockgrpp;
	ext4_grpblk_t offsetp;
	ext4_fsblk_t store_blocknr;
	ext4_get_group_no_and_offset(sb, blocknr, &blockgrpp, &offsetp);
	store_blocknr = ldfs_store_blocknr(sb, blockgrpp, offsetp);
	if (offset)
		*offset = ldfs_store_block_offset(offsetp);
	if (!ldfs_store_prefetched(sb, store_blocknr)) {
		ldfs_store_prefetch(sb, blockgrpp);
	}
	return store_blocknr;
}

int ldfs_store_read(struct super_block * sb, ext4_fsblk_t blocknr, __u8 fp[], __u32 * cnt) {
	ext4_fsblk_t store_blocknr;
	__u32 offset;
	struct buffer_head * bh;
	struct ldfs_s_entry * entry;

	store_blocknr = ldfs_get_store_position(sb, blocknr, &offset);
	bh = sb_getblk(sb, store_blocknr);
	if (!bh_uptodate_or_lock(bh))
		if (bh_submit_read(bh) < 0)
			BUG();

	entry = ((struct ldfs_s_entry *) bh->b_data) + offset;
	if (fp)
		memcpy(fp, entry->fp, EXT4_SB(sb)->s_fp_info->len);
	if (cnt)
		*cnt = entry->count;

	brelse(bh);
	return 0;
}

int ldfs_store_write(handle_t * handle, struct inode * inode, ext4_fsblk_t blocknr, __u8 fp[], __u32 cnt) {
	struct super_block * sb = inode->i_sb;
	ext4_fsblk_t store_blocknr;
	__u32 offset;
	struct buffer_head * bh;
	struct ldfs_s_entry * entry;

	store_blocknr = ldfs_get_store_position(sb, blocknr, &offset);
	bh = sb_getblk(sb, store_blocknr);
	if (!bh_uptodate_or_lock(bh))
		if (bh_submit_read(bh) < 0)
			BUG();

	entry = ((struct ldfs_s_entry *) bh->b_data) + offset;
	ext4_journal_get_write_access(handle, bh);
	if (fp)
		memcpy(entry->fp, fp, EXT4_SB(sb)->s_fp_info->len);
	if (cnt)
		entry->count = cnt;
	ext4_handle_dirty_metadata(handle, inode, bh);
	brelse(bh);
	return 0;
}

int ldfs_store_fp_cmp(struct inode * inode, ext4_fsblk_t blocknr, __u8 fp[]) {
	struct super_block * sb = inode->i_sb;
	ext4_fsblk_t store_blocknr;
	__u32 offset;
	struct buffer_head * bh;
	struct ldfs_s_entry * entry;
	int ret;

	store_blocknr = ldfs_get_store_position(sb, blocknr, &offset);
	bh = sb_getblk(inode->i_sb, store_blocknr);
	if (!bh_uptodate_or_lock(bh))
		if (bh_submit_read(bh) < 0)
			BUG();

	entry = ((struct ldfs_s_entry *) bh->b_data) + offset;
	ret = memcmp(fp, entry->fp, EXT4_SB(sb)->s_fp_info->len);
	brelse(bh);
	return ret;
}

int ldfs_store_count_link(handle_t * handle, struct inode * inode, ext4_fsblk_t blocknr) {
	struct super_block * sb = inode->i_sb;
	ext4_fsblk_t store_blocknr;
	__u32 offset;
	struct buffer_head * bh;
	struct ldfs_s_entry * entry;

	store_blocknr = ldfs_get_store_position(sb, blocknr, &offset);
	bh = sb_getblk(sb, store_blocknr);
	if (!bh_uptodate_or_lock(bh))
		if (bh_submit_read(bh) < 0)
			BUG();

	entry = ((struct ldfs_s_entry *) bh->b_data) + offset;
	ext4_journal_get_write_access(handle, bh);
	entry->count++;
	ext4_handle_dirty_metadata(handle, inode, bh);
	brelse(bh);
	return 0;
}

int ldfs_store_count_unlink(handle_t * handle, struct inode * inode, ext4_fsblk_t blocknr, __u8 fp[], __u32 * cnt) {
	struct super_block * sb = inode->i_sb;
	ext4_fsblk_t store_blocknr;
	__u32 offset;
	struct buffer_head * bh;
	struct ldfs_s_entry * entry;

	store_blocknr = ldfs_get_store_position(sb, blocknr, &offset);
	bh = sb_getblk(sb, store_blocknr);
	if (!bh_uptodate_or_lock(bh))
		if (bh_submit_read(bh) < 0)
			BUG();

	entry = ((struct ldfs_s_entry *) bh->b_data) + offset;
	if (entry->count == 0) {
		brelse(bh);
		return -1;
	}
	if (fp) {
		memcpy(fp, entry->fp, EXT4_SB(sb)->s_fp_info->len);
	}
	ext4_journal_get_write_access(handle, bh);
	entry->count--;
	if (cnt) {
		*cnt = entry->count;
	}
	ext4_handle_dirty_metadata(handle, inode, bh);
	brelse(bh);
	return 0;
}

int ldfs_store_fetch(struct super_block * sb, ext4_fsblk_t start_blocknr, __u32 cnt,
		insert_fp_t insert_fp, struct ldfs_lcache_group * cache_group) {
	struct buffer_head * bh;
	struct ldfs_s_entry * entry;
	ext4_fsblk_t store_start_blocknr, store_end_blocknr, cur_blocknr, processing_blocknr =
			start_blocknr;
	__u32 store_start_offset, store_end_offset, offset1, offset2;
	__u32 j;
	store_start_blocknr = ldfs_get_store_position(sb, start_blocknr, &store_start_offset);
	store_end_blocknr = ldfs_get_store_position(sb, start_blocknr + cnt - 1, &store_end_offset);
	offset1 = store_start_offset;
	offset2 = LDFS_S_CNT;

	for (cur_blocknr = store_start_blocknr; cur_blocknr <= store_end_blocknr; cur_blocknr++) {
		bh = sb_getblk(sb, cur_blocknr);
		if (!bh_uptodate_or_lock(bh))
			if (bh_submit_read(bh) < 0)
				BUG();

		if (cur_blocknr == store_end_blocknr)
			offset2 = store_end_offset + 1;
		for (j = offset1; j < offset2; j++) {
			entry = ((struct ldfs_s_entry *) bh->b_data) + j;
			insert_fp(sb, cache_group, processing_blocknr, entry->fp, entry->count);
			processing_blocknr++;
		}
		brelse(bh);
		offset1 = 0;
	}
	return 0;

}

int ldfs_store_mount(struct super_block * sb) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
	struct buffer_head * bh;
	struct ldfs_s_entry * entry;
	ext4_group_t group;
	ext4_fsblk_t store_blocknr, cur_blocknr = 0;
	__u32 store_size = EXT4_BLOCKS_PER_GROUP(sb) / LDFS_S_CNT * LDFS_STORE_COMPACT;
	int i, j;

	for (group = 0; group < sbi->s_groups_count; group += LDFS_STORE_COMPACT) {
		store_blocknr = ldfs_store_blocknr(sb, group, 0);
		for (i = 0; i < store_size; i++) {
			bh = sb_getblk(sb, store_blocknr + i);
			if (buffer_uptodate(bh)) {
				brelse(bh);
				continue;
			}
			lock_buffer(bh);
			get_bh(bh);
			bh->b_end_io = end_buffer_read_sync;
			submit_bh(READ, bh);
			brelse(bh);
		}
		for (i = 0; i < store_size; i++) {
			bh = sb_getblk(sb, store_blocknr + i);
			if (!buffer_uptodate(bh)) {
				wait_on_buffer(bh);
			}
			for (j = 0; j < LDFS_S_CNT; j++) {
				entry = ((struct ldfs_s_entry *) bh->b_data) + j;
				if (entry->count) {
					ldfs_filter_insert_entry(sb, entry->fp, cur_blocknr);
				}
				cur_blocknr++;
			}
			brelse(bh);
		}
	}
	return 0;
}

